//----------------------------------------------------------------------------
// DNA to RNA converter
//   Usage:
//     d2r > endo.rna
//     d2r my.prefix > prefix+endo.rna
//----------------------------------------------------------------------------

#include <ext/rope>
#include <vector>
#include <iostream>
#include <fstream>
#include <iterator>
#include <string>
#include <algorithm>
#include <boost/function.hpp>
#include <boost/timer.hpp>
using namespace std;

typedef __gnu_cxx::rope<char>  String;
typedef String::const_iterator StrItr;
typedef std::string            Buffer;

#define DebugOut(expr) 
//#define DebugOut(expr) (cerr<<expr)
#define STOP do{Buffer b; getline(cin, b);}while(0)

// Output ////////////////////////////////////////////////////////////////////

String dna;
StrItr dna_p, dna_e;

void rna()
{
	for(int k=min(7, dna_e-dna_p); k--; ++dna_p)
		cout.put(*dna_p);
	cout.put('\n');
}

void finish()
{
	cout << flush;
	exit(0);
}

// Pattern ///////////////////////////////////////////////////////////////////

typedef vector<String>                                  Env;
typedef boost::function<bool(int&,Env&,vector<int>&)> PItem;
typedef vector<PItem>                               Pattern;

int nat()
{
	for(int n=0, t=1; dna_p!=dna_e; t*=2)
	{
		char c = *dna_p; ++dna_p;
		if( c=='P' ) return n;
		if( c=='C' ) n += t;
	}
	finish();
}

Buffer consts()
{
	Buffer s;
	while( dna_p != dna_e ) {
		char c = *dna_p;
		     if( c=='C' ) { ++dna_p; s+='I'; }
		else if( c=='F' ) { ++dna_p; s+='C'; }
		else if( c=='P' ) { ++dna_p; s+='F'; }
		else if( dna_e-dna_p>=2 && *(dna_p+1)=='C' ) { dna_p+=2; s+='P'; }
		else break;
	}
	return s;
}

struct PItem_b
{
	Buffer b;
	PItem_b( Buffer& b_ ) { b.swap(b_); DebugOut(b); }

	bool operator()(int& i, Env& e, vector<int>& c)
	{
		if( dna.size() < b.size() )
			return false;
		for(int j=0; j!=b.size(); ++j,++i)
			if( dna[i]!=b[j] )
				return false;
		return true;
	}
};

struct PItem_n
{
	int n;
	PItem_n() : n(nat()) { DebugOut('!'<<n); }

	bool operator()(int& i, Env& e, vector<int>& c)
	{
		return (i+=n) <= dna.size();
	}
};

struct PItem_s
{
	Buffer s;
	PItem_s() : s(consts()) { DebugOut("?'"<<s<<"'"); }

	bool operator()(int& i, Env& e, vector<int>& c)
	{
		int n = dna.find(s.c_str(), i);
		if( n == String::npos )
			return false;
		i = n+s.size();
		return true;
	}
};

struct PItem_l
{
	PItem_l() { DebugOut('('); }

	bool operator()(int& i, Env& e, vector<int>& c)
	{
		c.push_back(i);
		return true;
	}
};

struct PItem_r
{
	PItem_r() { DebugOut(')'); }

	bool operator()(int& i, Env& e, vector<int>& c)
	{
		e.push_back( dna.substr(c.back(), i-c.back()) );
		c.pop_back();
		return true;
	}
};

Pattern pattern()
{
	Pattern p;
	Buffer  b;
	int   lvl = 0;
	enum {U=0x0100, B=0x0200, UI=0x0300, BI=0x0400, II=0x0500} s = U;

	while( dna_p != dna_e )
		switch( s | *dna_p++ )
		{
		case  U|'I': s=UI; break;
		case  B|'I': s=BI; break;
		case  U|'C':
		case  B|'C': b+='I'; s=B; break; // C
		case  U|'F':
		case  B|'F': b+='C'; s=B; break; // F
		case  U|'P':
		case  B|'P': b+='F'; s=B; break; // P
		case UI|'C':
		case BI|'C': b+='P'; s=B; break; // IC
		case BI|'P': p.push_back(PItem_b(b));
		case UI|'P': p.push_back(PItem_n( )); s=U; break; // IP
		case BI|'F': p.push_back(PItem_b(b));
		case UI|'F': if(dna_p!=dna_e)++dna_p;p.push_back(PItem_s()); s=U; break; // IF
		case BI|'I': p.push_back(PItem_b(b));
		case UI|'I': s=II; break;
		case II|'P': lvl++; p.push_back(PItem_l()); s=U; break; // IIP
		case II|'C': 
		case II|'F': if(lvl--)p.push_back(PItem_r());else return p; s=U; break; // IIC,IIF
		case II|'I': rna(); s=U; break; // III
		}

	finish();
}

// Template //////////////////////////////////////////////////////////////////

typedef boost::function<void(String& r, const Env& e)> TItem;
typedef vector<TItem>                               Template;

struct TItem_b
{
	Buffer b;
	TItem_b( Buffer& b_ ) { b.swap(b_); DebugOut(b); }

	void operator()(String& r, const Env& e) { r+=String(b.c_str(),b.size()); }
};

struct TItem_ln
{
	int l, n;
	TItem_ln() : l(nat()), n(nat()) {
		DebugOut('$'<<n);
		if(l) DebugOut('['<<l<<']');
	}

	void operator()(String& r, const Env& e){ r+=protect(l,n<e.size()?e[n]:String()); }

	static String protect(int ll, String d)
	{
		while(ll--)
			d = quote(d);
		return d;
	}

	static String quote(String d)
	{
		Buffer s;
		for(int i=0; i!=d.size(); ++i)
			switch( d[i] ) {
			case 'I': s += 'C'; break;
			case 'C': s += 'F'; break;
			case 'F': s += 'P'; break;
			case 'P': s += "IC"; break;
			}
		return s.c_str();
	}
};

struct TItem_n
{
	int n;
	TItem_n() : n(nat()) { DebugOut('|'<<n<<'|'); }

	void operator()(String& r, const Env& e){ asnat(r, n<e.size()?e[n].size():0); }

	static void asnat(String& r, int nn)
	{
		for(;nn;nn/=2)
			r += (nn%2 ? 'C' : 'I');
		r += 'P';
	}
};

// Parsing Templates /////////////////////////////////////////////////////////

Template template_()
{
	Template t;
	Buffer   b;
	enum {U=0x0100, B=0x0200, UI=0x0300, BI=0x0400, II=0x0500} s = U;

	while( dna_p != dna_e )
		switch( s | *dna_p++ )
		{
		case  U|'I': s=UI; break;
		case  B|'I': s=BI; break;
		case  U|'C':
		case  B|'C': b+='I'; s=B; break; // C
		case  U|'F':
		case  B|'F': b+='C'; s=B; break; // F
		case  U|'P':
		case  B|'P': b+='F'; s=B; break; // P
		case UI|'C':
		case BI|'C': b+='P'; s=B; break; // IC
		case BI|'P':
		case BI|'F': t.push_back(TItem_b(b));
		case UI|'P':
		case UI|'F': t.push_back(TItem_ln()); s=U; break; // IF,IP
		case BI|'I': t.push_back(TItem_b(b));
		case UI|'I': s=II; break;
		case II|'P': t.push_back(TItem_n( )); s=U; break; // IIP
		case II|'C':
		case II|'F': return t; // IIC, IIF
		case II|'I': rna(); s=U; break; // III
		}

	finish();
}

// Executing Templates ///////////////////////////////////////////////////////

void replace( const Template& tpl, const Env& e )
{
	String r;
	for(int i=0; i!=tpl.size(); ++i)
		tpl[i](r, e);
	dna = r + dna;
}

// Executing Patterns ////////////////////////////////////////////////////////

void matchreplace( const Pattern& pat, const Template& t )
{
	int i = 0;
	Env e;
	vector<int> c;

	for(int j=0; j!=pat.size(); ++j)
		if( !pat[j](i,e,c) )
			return;
	dna = dna.substr(i, dna.size()-i);
	replace(t, e);
}

// main //////////////////////////////////////////////////////////////////////

int main( int argc, const char* argv[] )
{
	// read
	if( argc>=2 )
	{
		ifstream fin(argv[1]);
		if(!fin) { cerr << "prefix file '" << argv[1] << "' cannot be opend" << endl; finish(); }
		istreambuf_iterator<char> fb(fin), fe;
		vector<char> buf( fb, fe );
		dna = String( &buf[0], buf.size() );
	}
	{
		ifstream fin("Task/endo.dna");
		istreambuf_iterator<char> fb(fin), fe;
		vector<char> buf( fb, fe );
		dna += String( &buf[0], buf.size() );
	}

	// exec
	boost::timer tim;
	for(int ite=1;;++ite)
	{
		dna_p = dna.begin();
		dna_e = dna.end();

		Pattern  p = pattern();   DebugOut(endl);
		Template t = template_(); DebugOut(endl);
		dna = dna.substr(dna_p, dna_e);
		matchreplace(p, t);

		if( ite%10000==0 )
			cerr << "Iteration " << ite << " finished (" << tim.elapsed() << " sec)" << endl;
	}
}
